import logging
from contextlib import asynccontextmanager

from fastapi import FastAPI, HTTPException
from starlette.responses import HTMLResponse

from eidolon_browser_service.api import PageInfo, EvaluateRequest, EvaluateInfo, NavigateRequest
from eidolon_browser_service.service import BrowserService

logger = logging.getLogger("browser-service")
browser_service = BrowserService()


@asynccontextmanager
async def lifespan(app):
    await browser_service.start()
    yield
    await browser_service.stop()


app = FastAPI(lifespan=lifespan)


@app.post("/contexts/{context_id}/pages")
async def create_page(context_id: str):
    context = await browser_service.get_or_create_context(context_id)
    page = await context.create_page()
    return PageInfo(page_id=page.page_id, url=None)


@app.get("/contexts/{context_id}/pages")
async def list_pages(context_id: str):
    context = await browser_service.get_or_create_context(context_id)
    return {
        "pages": [
            PageInfo(page_id=page.page_id, url=page.page.url if page.page.url != "about:blank" else None)
            for page in context.pages.values()
        ]
    }


@app.delete("/contexts/{context_id}")
async def delete_context(context_id: str):
    await browser_service.delete_context(context_id)
    return {"status": "deleted"}


@app.post("/contexts/{context_id}/pages/{page_id}/evaluate")
async def evaluate_script(context_id: str, page_id: str, request: EvaluateRequest):
    context = await browser_service.get_or_create_context(context_id)
    page = context.get_page(page_id)
    try:
        result = await page.page.evaluate(request.script)
    except Exception as e:
        if logger.isEnabledFor(logging.DEBUG):
            logger.warning(f"Error executing script:\n{request.script}", exc_info=True)
        else:
            logger.warning("Error executing script")
        raise HTTPException(status_code=422, detail=f"{type(e).__name__}: {e}")
    return EvaluateInfo(result=str(result) if result is not None else None)


@app.post("/contexts/{context_id}/pages/{page_id}/navigate")
async def navigate(context_id: str, page_id: str, request: NavigateRequest):
    context = await browser_service.get_or_create_context(context_id)
    page = context.get_page(page_id)
    await page.page.goto(request.url)


@app.get("/contexts/{context_id}/pages/{page_id}/content", response_class=HTMLResponse)
async def get_page_content(
        context_id: str,
        page_id: str,
):
    context = await browser_service.get_or_create_context(context_id)
    page = context.get_page(page_id)
    return await page.page.content()



from typing import Dict, Optional
import uuid

from fastapi import HTTPException
from playwright.async_api import Page


class ManagedPage:
    page_id: str
    page: Page

    def __init__(self, page_id: str, page: Page):
        self.page_id = page_id
        self.page = page


class BrowserContext:
    def __init__(self, context, browser):
        self.context = context
        self.browser = browser
        self.pages: Dict[str, ManagedPage] = {}

    async def create_page(self) -> ManagedPage:
        page = await self.context.new_page()
        page_id = str(uuid.uuid4())
        managed_page = ManagedPage(page_id, page)
        self.pages[page_id] = managed_page
        return managed_page

    async def cleanup(self):
        await self.context.close()
        await self.browser.close()

    def get_page(self, page_id: str) -> ManagedPage:
        page = self.pages.get(page_id)
        if not page:
            raise HTTPException(status_code=404, detail="Page not found")
        return page



from typing import Dict
from playwright.async_api import async_playwright
from .browser import BrowserContext


class BrowserService:
    def __init__(self):
        self.contexts: Dict[str, BrowserContext] = {}
        self.playwright = None
        self.browser = None

    async def start(self):
        self.playwright = await async_playwright().start()
        self.browser = await self.playwright.chromium.launch()

    async def stop(self):
        if self.playwright:
            await self.playwright.stop()
        self.playwright = None
        self.browser = None

    async def get_or_create_context(self, context_id: str) -> BrowserContext:
        if context_id not in self.contexts:
            context = await self.browser.new_context()
            self.contexts[context_id] = BrowserContext(context, self.browser)
        return self.contexts[context_id]

    async def delete_context(self, context_id: str):
        if context_id in self.contexts:
            await self.contexts[context_id].cleanup()
            del self.contexts[context_id]



from typing import Optional

from pydantic import BaseModel


class PageInfo(BaseModel):
    page_id: str
    url: Optional[str]


class EvaluateRequest(BaseModel):
    script: str


class EvaluateInfo(BaseModel):
    result: Optional[str]


class NavigateRequest(BaseModel):
    url: str



from __future__ import annotations

import os
from typing import List
from urllib.parse import urljoin

import httpx
from pydantic import BaseModel

from eidolon_browser_service.api import PageInfo, EvaluateInfo, NavigateRequest


class BrowserError(Exception):
    def __init__(self, message: str, status_code: int = None, response_body: str = None,
                 original_error: Exception = None):
        super().__init__(message)
        self.status_code = status_code
        self.response_body = response_body
        self.original_error = original_error


class PageNotFoundError(BrowserError):
    pass


class EvaluationError(BrowserError):
    pass


async def _handle_response_error(e: httpx.HTTPStatusError, context: str) -> None:
    if e.response.status_code == 404:
        raise PageNotFoundError(
            f"{context}: resource not found",
            status_code=e.response.status_code,
            response_body=e.response.text,
            original_error=e
        )
    raise BrowserError(
        f"{context}: {str(e)}",
        status_code=e.response.status_code,
        response_body=e.response.text,
        original_error=e
    )


class Page(PageInfo):
    location: str
    context_id: str

    async def evaluate(self, script: str) -> EvaluateInfo:
        try:
            async with httpx.AsyncClient() as client:
                response = await client.post(
                    urljoin(self.location, f"/contexts/{self.context_id}/pages/{self.page_id}/evaluate"),
                    json={"script": script}
                )
                response.raise_for_status()
                return EvaluateInfo.model_validate(response.json())
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 422:
                raise EvaluationError(
                    e.response.json().get('detail', e.response.text),
                    status_code=e.response.status_code,
                    response_body=e.response.text,
                    original_error=e
                )
            await _handle_response_error(e, "Evaluation failed")

    async def navigate(self, url: str) -> Page:
        try:
            async with httpx.AsyncClient() as client:
                response = await client.post(
                    urljoin(self.location, f"/contexts/{self.context_id}/pages/{self.page_id}/navigate"),
                    json=NavigateRequest(url=url).model_dump(),
                )
                response.raise_for_status()
                return Page(
                    location=self.location,
                    context_id=self.context_id,
                    page_id=self.page_id,
                    url=url,
                )
        except httpx.HTTPStatusError as e:
            await _handle_response_error(e, "Failed to navigate")

    async def get_content(self) -> str:
        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(
                    urljoin(self.location, f"/contexts/{self.context_id}/pages/{self.page_id}/content")
                )
                response.raise_for_status()
                return response.text
        except httpx.HTTPStatusError as e:
            await _handle_response_error(e, "Failed to get content")


class Context(BaseModel):
    location: str
    context_id: str

    async def create_page(self) -> Page:
        try:
            async with httpx.AsyncClient() as client:
                response = await client.post(urljoin(self.location, f"/contexts/{self.context_id}/pages"))
                response.raise_for_status()
                return Page(
                    location=self.location,
                    context_id=self.context_id,
                    **response.json()
                )
        except httpx.HTTPStatusError as e:
            await _handle_response_error(e, "Failed to create page")

    async def list_pages(self) -> List[Page]:
        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(
                    urljoin(self.location, f"/contexts/{self.context_id}/pages")
                )
                response.raise_for_status()
                return [
                    Page(
                        location=self.location,
                        context_id=self.context_id,
                        **PageInfo(**page).model_dump(),
                    )
                    for page in response.json()["pages"]
                ]
        except httpx.HTTPStatusError as e:
            await _handle_response_error(e, "Failed to list pages")

    async def delete(self):
        try:
            async with httpx.AsyncClient() as client:
                response = await client.delete(
                    urljoin(self.location, f"/contexts/{self.context_id}")
                )
                response.raise_for_status()
        except httpx.HTTPStatusError as e:
            await _handle_response_error(e, "Failed to delete context")


class Browser(BaseModel):
    location: str = os.environ.get("BROWSER_SERVICE_URL", "http://localhost:7468")

    def context(self, context_id: str) -> Context:
        return Context(location=self.location, context_id=context_id)
